let output = ref Strtree.create

let do_output () =
  let remaining_time = Unix.alarm 0 in (* [ES] be careful to avoid race *)
  Strtree.output stdout !output

let timeout signal_number =
  if Debug.level >= 1 then
    Format.eprintf "signal %d =? %d@." signal_number Sys.sigalrm;
  (* assert (signal_number = Sys.sigalrm); *)
  if Debug.level >= 0 then Format.eprintf "timed out@.";
  do_output ();
  exit 0

let input_all () =
  let bufsize = 6000000 in
  let buf = String.create bufsize in
  let rec input_all_acc pos =
    let len = input stdin buf pos (bufsize - pos) in
    if len = 0 then pos else
    input_all_acc (pos + len) in
  let pos = input_all_acc 0 in
  String.sub buf 0 pos

let update_output (taglen, answer) =
  let prevmask = Unix.sigprocmask Unix.SIG_BLOCK [Sys.sigalrm] in
  let anslen, outlen = Strtree.length answer, Strtree.length !output in
  if taglen < !Search.current_taglen then begin
    if anslen < outlen then begin
      if Debug.level >= 2 then
	Format.eprintf "!length: %d(%d) -> %d(%d): %s@." 
	  outlen !Search.current_taglen 
	  anslen taglen
	  (Strtree.to_string answer)
      else if Debug.level >= 1 then
	Format.eprintf "!length: %d(%d) -> %d(%d)@." 
	  outlen !Search.current_taglen
	  anslen taglen;
      output := answer;
    end else begin 
      if Debug.level >= 2 then
	Format.eprintf "?length: %d(%d) -> %d(%d): %s@." 
	  outlen !Search.current_taglen 
	  anslen taglen
	  (Strtree.to_string answer)
      else if Debug.level >= 1 then
	Format.eprintf "?length: %d(%d) -> %d(%d)@." 
	  outlen !Search.current_taglen
	  anslen taglen;
    end;
    Search.current_taglen := taglen;
  end else
      if Debug.level >= 2 then
	Format.eprintf "length: %d(%d) -> %d(%d): %s@." 
	  outlen !Search.current_taglen 
	  anslen taglen
	  (Strtree.to_string answer)
      else if Debug.level >= 1 then
	Format.eprintf "length: %d(%d) -> %d(%d)@." 
	  outlen !Search.current_taglen
	  anslen taglen;
  ignore (Unix.sigprocmask Unix.SIG_SETMASK prevmask)

let _ =
  let input = input_all () in
  if Debug.level >= 1 then Format.eprintf "input done@.";
  output := Strtree.from_string input;

  Sys.set_signal Sys.sigalrm (Sys.Signal_handle timeout);
  let time_limit = try int_of_string (Unix.getenv "TLIMIT") with _ -> max_int in
  let previous_alarm = Unix.alarm (time_limit - 10) in
    (* [ES] margin for the time of outputting itself *)
  assert (previous_alarm = 0);

  let abssyn = ReadInput.tree_of_string input in
  if Debug.level >= 1 then Format.eprintf "parsing done@.";
  let pp1 = Pp.f abssyn in 
  if Debug.level >= 1 then Format.eprintf "pp1 done@.";
  update_output (max_int - 1, pp1);

  let sem = Semantics.f Semantics.default_attr abssyn in
  if Debug.level >= 1 then Format.eprintf "eval done@.";
  let sem' = UnifySpace.f sem in
  if Debug.level >= 1 then Format.eprintf "space unification done@.";
  let (taglen, str) = UnifySpace.pp sem' in
  if Debug.level >= 1 then Format.eprintf "pp2 done@.";
  update_output (taglen, str);

  Search.current_taglen := taglen;
  Search.update_output := update_output;
  Search.f sem';

  do_output ();
  exit 0
