%{
open AbsSyn

let tt_stack = ref []
let tt_flag = ref false
let tt_save b = 
  tt_stack := !tt_flag :: !tt_stack;
  tt_flag := b
let tt_restore () = 
  tt_flag := List.hd !tt_stack;
  tt_stack := List.tl !tt_stack

%}

%token ST_B END_B
%token ST_EM END_EM
%token ST_I END_I
%token ST_PL END_PL
%token ST_S END_S
%token ST_TT END_TT
%token ST_U END_U
%token <int> ST_SIZE END_SIZE
%token <char> ST_COLOR END_COLOR
%token <string> TEXT SPACE
%token EOF
%type <AbsSyn.document> input_text
%start input_text

%%
input_text:
  document EOF { assert (!tt_stack == []); $1 }

document:
  /* empty */ { [] }
| document_elem document { $1 :: $2 }

tt_set:
    /* empty */ { tt_save true }
tt_reset:
    /* empty */ { tt_save false }

document_elem:
  TEXT { Text $1 }
| SPACE { Space (if !tt_flag then $1 else " ") }
| ST_B document END_B { Bold $2 }
| ST_EM document END_EM { Emph $2 }
| ST_I document END_I { Italic $2 }
| ST_PL tt_reset document END_PL { tt_restore (); Plain $3 }
| ST_S document END_S { Strong $2 }
| ST_TT tt_set document END_TT { tt_restore (); Typewriter $3 }
| ST_U document END_U { Underline $2 }
| ST_SIZE document END_SIZE
    { if $1 = $3 then Size($1,$2) else failwith "unmatched size tag" }
| ST_COLOR document END_COLOR
    { if $1 = $3 then Color($1,$2) else failwith "unmatched color tag" }

