type ems_t = No | Em | S

let emph_ems = function
  | No -> Em
  | Em -> No
  | S -> S

type attr = 
    { b : bool;
      ems : ems_t;
      i : bool;
      tt : bool;
      u : int; (* 0..3 *)
      size : int option; (* 0..9 or unknown *)
      color : char option; (* {r,g,b,c,m,y,k,w} or unknown *) }

let default_attr =
  { b = false;
    ems = No;
    i = false;
    tt = false;
    u = 0;
    size = None;
    color = None }

type spc_attr = 
    { spc_tt : bool;
      spc_ucolor : (int (* 1..3 *) * char option) option; (* None means u = 0 *)
      spc_size : int option }

type t =
    (* invariant: these strings cannot be null *)
  | Text of attr * string
  | Space of spc_attr * string
      (* invariant: if attr.spc_tt = false, then string = " " *)

let spc_ctx attr =
  { spc_tt = attr.tt;
    spc_ucolor =
    if attr.u = 0 then None else
    Some(attr.u, attr.color);
    spc_size = attr.size }

let un_spc_ctx attr spc_attr =
  (* copies attributes b, ems, i from the current context
     copies color if not underlined *)
  let (u, c) =
    match spc_attr.spc_ucolor with
    | None -> (0, attr.color)
    | Some(i, c) -> (i, c) in
  { b = attr.b;
    ems = attr.ems;
    i = attr.i;
    tt = spc_attr.spc_tt;
    u = u;
    size = spc_attr.spc_size;
    color = c }

let rec f prevattr = function
  | [] -> []
  | AbsSyn.Text s :: rest -> Text (prevattr, s) :: f prevattr rest
  | AbsSyn.Space s :: rest ->
      let thisattr = spc_ctx prevattr in
      Space (thisattr, if thisattr.spc_tt then s else " ") :: f prevattr rest
  | AbsSyn.Bold d :: rest ->
      let newattr = { prevattr with b = true } in
      f newattr d @ f prevattr rest
  | AbsSyn.Emph d :: rest ->
      let newattr = { prevattr with ems = emph_ems prevattr.ems } in
      f newattr d @ f prevattr rest
  | AbsSyn.Italic d :: rest ->
      let newattr = { prevattr with i = true } in
      f newattr d @ f prevattr rest
  | AbsSyn.Plain d :: rest ->
      let newattr = { prevattr with
		      u = 0;
		      ems = No;
		      i = false;
		      b = false;
		      tt = false } in
      f newattr d @ f prevattr rest
  | AbsSyn.Strong d :: rest ->
      let newattr = { prevattr with ems = S; } in
      f newattr d @ f prevattr rest
  | AbsSyn.Typewriter d :: rest ->
      let newattr = { prevattr with tt = true } in
      f newattr d @ f prevattr rest
  | AbsSyn.Underline d :: rest ->
      let newattr = { prevattr with u = min 3 (prevattr.u + 1) } in
      f newattr d @ f prevattr rest
  | AbsSyn.Size (i, d) :: rest ->
      let newattr = { prevattr with size = Some i } in
      f newattr d @ f prevattr rest
  | AbsSyn.Color (c, d) :: rest ->
      let newattr = { prevattr with color = Some c } in
      f newattr d @ f prevattr rest
