type tags = 
    Bold | Emph | Italic | Plain | Strong | Typewriter 
  | Underline | Size of int | Color of char

let tag_of_int = 
  [| Size 0; Size 1; Size 2; Size 3; Size 4; Size 5; Size 6; Size 7; Size 8; Size 9;
     Color 'r'; Color 'g'; Color 'b'; Color 'c'; Color 'm'; Color 'y'; Color 'k'; 
     Color 'w'; Emph; Italic; Plain; Strong; Typewriter; Underline|]

let number_of_tags = Array.length tag_of_int

let opening_of_tags = function
    Bold -> "<B>"
  | Emph -> "<EM>"
  | Italic -> "<I>"
  | Plain -> "<PL>"
  | Strong -> "<S>"
  | Typewriter -> "<TT>"
  | Underline -> "<U>"
  | Size n -> Printf.sprintf "<%s>" (string_of_int n)
  | Color c -> Printf.sprintf "<%c>" c

let closing_of_tags = function
    Bold -> "</B>"
  | Emph -> "</EM>"
  | Italic -> "</I>"
  | Plain -> "</PL>"
  | Strong -> "</S>"
  | Typewriter -> "</TT>"
  | Underline -> "</U>"
  | Size n -> Printf.sprintf "</%s>" (string_of_int n)
  | Color c -> Printf.sprintf "</%c>" c

let generate_text () = 
  (if Random.int 10 = 0 then " " else "") ^
  "hoge" ^ (if Random.int 10 = 0 then " " else "")

let rec generate_sml stack stack_length sml_length =
  (* stack_length must be the length of stack *)
  if (sml_length <= max 0 stack_length) then 
    List.iter (fun x -> print_string (closing_of_tags x)) stack
  else
    let r = Random.int (number_of_tags * (List.length stack / 2 + 1)) in
    let hoge = generate_text () in
    let length_hoge = String.length hoge in
    if r >= number_of_tags then 
      (* close the current tag on top of the stack *)
      let hd_stack :: tl_stack = stack in
      let ctag = closing_of_tags hd_stack in
      let rest_sml_length = sml_length - (String.length ctag + length_hoge) in
      begin
	print_string ctag;
	print_string hoge;
	generate_sml tl_stack (stack_length - String.length ctag) rest_sml_length 
      end
    else 
      let otag = opening_of_tags tag_of_int.(r) in
      let rest_sml_length = sml_length - (String.length otag + length_hoge) in
      begin
	print_string otag;
	print_string hoge;
	generate_sml (tag_of_int.(r) :: stack) 
	  (stack_length + String.length otag + 1) rest_sml_length
      end

let main =
  let sml_length = int_of_string Sys.argv.(1) in
  Random.self_init ();
  generate_sml [] 0 sml_length;
  exit 0;
  ()


