open Semantics
open Strtree

let rec f = function
  | (Text(a1, s1) as txt) :: ((Text(a2, s2) :: rest) as sem) ->
      if a1 <> a2 then txt :: f sem else
      f(Text(a1, s1 ^ s2) :: rest)
  | (Text(a1, s1) as txt) :: ((Space(a2, s2) :: rest) as sem) ->
      if spc_ctx a1 <> a2 then txt :: f sem else
      if (not a2.spc_tt) && s1.[String.length s1 - 1] = ' ' then
	f(Text(a1, s1) :: rest)
      else
	f(Text(a1, s1 ^ s2) :: rest)
  | (Space(a1, s1) as spc) :: ((Text(a2, s2) :: rest) as sem) ->
      if a1 <> spc_ctx a2 then spc :: f sem else
      if (not a1.spc_tt) && s2.[0] = ' ' then
	f(Text(a2, s2) :: rest)
      else
	f(Text(a2, s1 ^ s2) :: rest)
  | (Space(a1, s1) as spc) :: ((Space(a2, s2) :: rest) as sem) ->
      if a1 <> a2 then spc :: f sem else
      if a1.spc_tt then
	f(Space(a1, s1 ^ s2) :: rest)
      else
	f(Space(a1, " ") :: rest)
  | [txt_or_spc]-> [txt_or_spc]
  | [] -> []

(* returns tag length and string *)
let rec pp (acc_taglen, acc_s) rest =
  match rest with
  | [] -> acc_taglen, acc_s
  | Text(a, s) :: rest ->
      let taglen = 0 in
      let s = from_string s in
      let taglen, s =
	match a.color with
	| None -> taglen, s
	| Some x ->
	    taglen + 7,
	    concat2 [from_string (Printf.sprintf "<%c>" x);
		     s;
		     from_string (Printf.sprintf "</%c>" x)] in
      let taglen, s = 
	match a.size with
	| None -> taglen, s
	| Some x ->
	    taglen + 7,
	    concat2 [from_string (Printf.sprintf "<%d>" x);
		     s;
		     from_string (Printf.sprintf "</%d>" x)] in
      let taglen, s = 
	match a.u with
	| 0 -> taglen, s
	| 1 -> taglen + 7, concat2 [from_string "<U>"; s; from_string "</U>"]
	| 2 -> taglen + 14, concat2 [from_string "<U><U>"; s; from_string "</U></U>"]
	| 3 -> taglen + 21, concat2 [from_string "<U><U><U>"; s; from_string "</U></U></U>"]
	| _ -> assert false in
      let taglen, s = if a.tt then taglen + 9, concat2 [from_string "<TT>"; s; from_string "</TT>"] else taglen, s in
      let taglen, s =
	match a.ems with
	| S -> taglen + 7, concat2 [from_string "<S>"; s; from_string "</S>"]
	| Em -> taglen + 9, concat2 [from_string "<EM>"; s; from_string "</EM>"]
	| No -> taglen, s in
      let taglen, s = if a.i then taglen + 7, concat2 [from_string "<I>"; s; from_string "</I>"] else taglen, s in
      let taglen, s = if a.b then taglen + 7, concat2 [from_string "<B>"; s; from_string "</B>"] else taglen, s in
      pp (acc_taglen + taglen, concat2 [acc_s; s]) rest
  | Space(a, s) :: rest ->
      let taglen = 0 in
      let s = from_string s in
      let taglen, s =
	match a.spc_ucolor with
	| None -> taglen, s
	| Some (i, copt) ->
	    let taglen, s = match copt with
	      None -> taglen, s
	    | Some c -> taglen + 7, concat2 [from_string (Printf.sprintf "<%c>" c);
					     s;
					     from_string (Printf.sprintf "</%c>" c)] in
	    match i with
	    | 0 -> taglen, s
	    | 1 -> taglen + 7, concat2 [from_string "<U>"; s; from_string "</U>"]
	    | 2 -> taglen + 14, concat2 [from_string "<U><U>"; s; from_string "</U></U>"]
	    | 3 -> taglen + 21, concat2 [from_string "<U><U><U>"; s; from_string "</U></U></U>"]
	    | _ -> assert false in
      let taglen, s = 
	match a.spc_size with
	| None -> taglen, s
	| Some x -> taglen + 7, concat2 [from_string (Printf.sprintf "<%d>" x);
					 s;
					 from_string (Printf.sprintf "</%d>" x)] in
      let taglen, s = if a.spc_tt then taglen + 9, concat2 [from_string "<TT>";
							    s;
							    from_string "</TT>"] else taglen, s in
      pp (acc_taglen + taglen, concat2 [acc_s; s]) rest

let pp rest = pp (0, create) rest
